#!/bin/sh
# flcheck.tcl \
exec tclsh "$0" ${1+"$@"}

# flcheck --
#
# This file implements package flcheck, a fusion logic modelchecker 
# using the CUDD BDD package
#
# Copyright (C) 2009-2015  Antonio Cau

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


#  -*- Mode: Tcl -*-


# Libraries needed
#

# Tk library
  package require Tk
# tile library
  package require tile
# fusion logic parser/lexer
  if {![info exist ::starkit::topdir]} {
    source fusion_logic.tcl
  } else {
     source $::starkit::topdir/fusion_logic.tcl
  }
# tcldot which is broken so we use dot directly
#load ./libtcldot.so.0.0.0

# Tile theme used by GUI
#
namespace eval tile::theme::strl {

    package provide tile::theme::strl 0.8

    variable colors
    array set colors {
	-frame  	"#eadaa5"
	-lighter	"#f5f5f5"
	-window	 	"#ffffff"
	-selectbg	"#ffff33"
	-selectfg	"#000000"
	-disabledfg	"#666666"
    }

    ttk::style theme create strl -settings {

	ttk::style configure . \
	    -borderwidth 	1 \
	    -background 	$colors(-frame) \
	    -fieldbackground	$colors(-window) \
	    -troughcolor	$colors(-lighter) \
	    -selectbackground	$colors(-selectbg) \
	    -selectforeground	$colors(-selectfg) \
	    ;
	ttk::style map . -foreground [list disabled $colors(-disabledfg)]

	## Buttons.
	#
	ttk::style configure TButton \
	    -width 0 -padding "1 1" -relief raised -shiftrelief 1 \
	    -highlightthickness 1 -highlightcolor $colors(-frame)

	ttk::style map TButton -relief {
	    {pressed !disabled} 	sunken
	    {active !disabled} 	raised
	} -highlightcolor {alternate black}


	ttk::style configure TMenubutton -relief raised -padding {10 2}
	ttk::style configure TRadiobutton -padding 1
	ttk::style configure TCheckbutton -padding 1

	## Toolbar buttons.
	#
	ttk::style configure Toolbutton \
	    -width 0 -relief flat -borderwidth 2 -padding 4 \
	    -background $colors(-frame) -foreground #000000 ;
	ttk::style map Toolbutton -background [list active $colors(-selectbg)] 
	ttk::style map Toolbutton -foreground [list active $colors(-selectfg)] 
	ttk::style map Toolbutton -relief {
	    disabled 	flat
	    selected	sunken  
	    pressed 	sunken  
	    active  	raised
	}

	## Entry widgets.
	#
	ttk::style configure TEntry \
	    -selectborderwidth 1 -padding 2 -insertwidth 2 -font TkTextFont
	ttk::style configure TCombobox \
	    -selectborderwidth 1 -padding 2 -insertwidth 2 -font TkTextFont

	## Notebooks.
	#
        ttk::style configure TNotebook -padding 0 -tabmargins {2 2 1 0}
        ttk::style configure TNotebook.Tab -padding {4 2}
        ttk::style map TNotebook.Tab -expand [list selected {2 2 1 0}]

	## Labelframes.
	#
	ttk::style configure TLabelframe -borderwidth 2 -relief groove

	## Scrollbars.
	#
        ttk::style configure TScrollbar \
	    -width 12 -arrowsize 12 
	
 
	ttk::style layout Vertical.TScrollbar {
	    Scrollbar.trough -children {
		Scrollbar.uparrow -side top
		Scrollbar.downarrow -side bottom
		Scrollbar.uparrow -side bottom
		Vertical.Scrollbar.thumb -side top -expand true -sticky ns
	    }
	}

	ttk::style layout Horizontal.TScrollbar {
	    Scrollbar.trough -children {
		Scrollbar.leftarrow -side left
		Scrollbar.rightarrow -side right
		Scrollbar.leftarrow -side right
		Horizontal.Scrollbar.thumb -side left -expand true -sticky we
	    }
	}
    }
}


# from http://wiki.tcl.tk/1109
# 
namespace eval history {
    proc add? {w} {
        variable $w
        variable n$w
        upvar 0 $w hist
        set s [set ::[$w cget -textvariable]]
        if {$s == ""} return
        if [string compare $s [lindex $hist end]] {
            lappend hist $s
            set n$w [llength $hist]
        }
    }
    proc move {w where} {
        variable $w
        variable n$w
        upvar 0 $w hist
        incr n$w $where
        if {[set n$w]<0} {set n$w 0}
        if {[set n$w]>=[llength $hist]+1} {
            set n$w [llength $hist]
        }
        set ::[$w cget -textvar] [lindex $hist [set n$w]]
    }
    proc for {type name args} {
        switch -- $type {
            ttk::entry {
                uplevel $type $name $args
                bind $name <Up> {history::move %W -1}
                bind $name <Down> {history::move %W 1}
                bind $name <Next> {history::move %W 99999}
                bind $name <Return> {history::add? %W}
                variable $name {}
                variable n$name 0
            }
            default {error "usage: history::for entry <w> <args>"}
        }
    }
 }




# bdd_tcl package
#

namespace eval ::bdd_tcl {
global c BDD 

# choose a bdd package
proc init_bdd {} {
  global BDD
  if [string equal $BDD "CUDD"] {
   if {![info exist ::starkit::topdir]} {
     source cudd.tcl
   } else {
     source $::starkit::topdir/cudd.tcl
   }
  } else {
   if [string equal $BDD "Buddy"] {
     if {![info exist ::starkit::topdir]} {
      source buddy.tcl 
     } else {
      source $::starkit::topdir/buddy.tcl
     }
   }
  }
}


#treelist routines
#

proc graphInit _g {upvar 1 $_g g; catch {unset g}; set g(nodes) {}}

proc addNode {_g node args} {
    upvar 1 $_g g
    set id [llength $g(nodes)]
    set g($id) [concat $node $args]
    lappend g(nodes) $id
    set id
 }

proc addEdge {_g from to args} {upvar 1 $_g g; set g($from,$to) $args}

proc isTerminal {_g id} {
    upvar 1 $_g g
    expr {[array names g $id,*]==""} ;# no edges leading out
 }

proc sons {_g id} {
    upvar 1 $_g g
    regsub -all $id, [array names g $id,*] "" res
    set res
 }

proc edges _g {upvar 1 $_g g; array names g *,*}

proc nodes _g {upvar 1 $_g g; set g(nodes)}

proc treelist2graph {L _g} {
    upvar 1 $_g g
    set from [addNode g [lindex $L 0]]
    foreach i [lrange $L 1 end] {
        set to [treelist2graph $i g]
        addEdge g $from $to
    }
    set from
 }

proc layout _g {
    upvar 1 $_g g
    set xt 0; set yt 0; set dx -60; set dy -30
    foreach id [lrevert [nodes g]] {
        if [isTerminal g $id] {
            set g(xy:$id) [list $xt $yt]
            incr xt $dx
        } else {
            set xs {}; set ys {}
            foreach i [sons g $id] {
                foreach {x y} $g(xy:$i) break
                lappend xs $x
                lappend ys $y
            }
            set x [expr {([max $xs]+[min $xs])/2}] ;# x:center between sons
            set y [expr {[min $ys]+$dy}]  ;# y: one layer above highest son
            set g(xy:$id) [list $x $y]
        }
    }
 }

proc draw {_g w} {
  global bddfont
    upvar 1 $_g g
    foreach i [edges g] {
        foreach {from to} [split $i ,] break
        eval $w create line $g(xy:$from) $g(xy:$to) -width 2 -tag edge
    }
    foreach i [nodes g] {
        set id [eval $w create text $g(xy:$i) -text [list $g($i)] -tag node]
        eval $w create rect [$w bbox $id] -fill white -outline white -tag node
        $w raise $id
    }
 }

proc max L {lindex [lsort -integer $L] end}

proc min L {lindex [lsort -integer $L] 0}

proc lrevert L {
    for {set res {}; set i [llength $L]} {$i>0} {#see loop} {
        lappend res [lindex $L [incr i -1]]
    }
    set res
 }


proc drawtree { tl tabtext} {
global c tab

    graphInit g
    treelist2graph [lindex $tl 0] g

    new_tab .top18 $tab $tabtext
    incr tab
    #puts "tab: $tab"
    layout g
    $c delete edge node
    draw g $c
    foreach {x y} [$c bbox all] break
    $c move all [expr 5-$x] [expr 5-$y]
    zoom $c 0.8
#    foreach {- - x1 y1} [$c bbox all] break
#    $c config -width [expr $x1+10] -height [expr $y1+10]

}



# reduce routines
#


#
# c(e) where e is a fusion expression
#
# L: tree list (fusion expression) 
#
proc  c_fe {L} {

   set root [lindex $L 0]

   switch $root {
       fetest { 
           return " { fetest {stnot true} } "
               }
       festep { 
           return " { festep {[lindex $L 1]} } "
               }
       fechop { 
                 set x [c_fe [lindex $L 1]]
                 set y [c_fe [lindex $L 2]]
                 return " { feor { fechop $x {[lindex $L 2]} } { fechop {[lindex $L 1]} $y} } "
               }
       feor   { 
                 set x [c_fe [lindex $L 1]]
                 set y [c_fe [lindex $L 2]]
                 return " { feor $x $y } "
               }
       feand  {
                 set x [lindex $L 1]
                 set y [c_fe [lindex $L 2]]
                 return " { feand $x $y } "
              }
       fechopstar { 
                     set y $L
                     set x [c_fe [lindex $L 1]]
           return " { fechop $x { $y } } "
               }
       default  { 
              if {[llength $L] == 1 } {
                  return " { $L } "
              } else {
                if {[llength $L] == 2 } {
                    set oper [lindex $L 0]
                    set x [c_fe [lindex $L 1]]
                    return " { $oper $x } "
                } else {
                   if {[llength $L] == 3 } {
                     set oper [lindex $L 0]
                     set x [c_fe [lindex $L 1]]
                     set y [c_fe [lindex $L 2]]
                     return " { $oper $x $y } "
                   } else {
                     puts "error in c_fe"
                   }
                }
              }
       }
   }
}

#
# is_state L 
# determine whether L is a state formula
#
# L : tree list (fusion logic formula)
#
proc is_state { L } {
  set root [lindex $L 0]
  #puts "is_state: root is $root"
  switch -regexp -- $root {
      stor   { 
                return 1
      }
      stand  {
                return 1
      }
      stnot  { 
                return 1
      }
      true    { 
                return 1
      }
      false    { 
                return 1
      }
      {r_[0-9]+} { 
                 return 1
      }
      {r_q[0-9]+}    { 
                 return 1
      }  
      {[A-Z]+[_A-Z]*} { 
                 return 1
      }
      flor    { 
          set x [is_state [lindex $L 1]]
          set y [is_state [lindex $L 2]]
          if { $x==1 && $y==1 } { return 1} else {return 0}
      }
      fland    { 
          set x [is_state [lindex $L 1]]
          set y [is_state [lindex $L 2]]
          if { $x==1 && $y==1 } { return 1} else {return 0}
      }
      flnot   { 
          return [is_state [lindex $L 1]]
      }
      default  {  
              return 0
      }
  }
}

##
## shift L k: L \uparrow k
## shift the subscripts of the dependent variables in L by k
##
## L: tree list (invariant)
## k: integer (>0)
##
#proc shift { L k } {

#  set root [lindex $L 0]
 
#  switch -regexp -- $root {
#      {r_[0-9]+} { 
#                   set j [string range $root 2 end]
#                   set jk [expr $k + $j]
#                   return " r_$jk "
#      }
#      
#      default { 
#                if {[llength $L] == 1 } {
#                  return " { $L } "
#                } else {
#                   if {[llength $L] == 2 } {
#                     set oper [lindex $L 0]
#                     set x [shift [lindex $L 1] $k]
#                     return " { $oper $x } "
#                   } else {
#                     if {[llength $L] == 3 } {
#                       set oper [lindex $L 0]
#                       set x [shift [lindex $L 1] $k]
#                       set y [shift [lindex $L 2] $k]
#                       return " { $oper $x $y } "
#                     } else {
#                       puts "error in shift"
#                     }
#                   }
#                }
#      }
#  }
#              
#}


##
## reduce_length L
## determine the number of distinct dependent variables in L
## 
## L: tree list (reduced fusion logic formula)
##
#proc reduce_length { L } {
#
#  set root [lindex $L 0]
#
#  switch $root {
#      Eqv { 
#            return 1
#      }
#      default { 
#                if {[llength $L] == 1 } {
#                  return 0
#                } else {
#                   if {[llength $L] == 2 } {
#                     set oper [lindex $L 0]
#                     set x [reduce_length [lindex $L 1]]
#                     return $x
#                   } else {
#                     if {[llength $L] == 3 } {
#                       set oper [lindex $L 0]
#                       set x [reduce_length [lindex $L 1]]
#                       set y [reduce_length [lindex $L 2]]
#                       set xy [expr $x + $y]
#                       return $xy
#                     } else {
#                       puts "error in reduce_length"
#                     }
#                   }
#                }
#      }
#  }
#
#}

##
## reduce L
## reduction function transforms a fusion logic formula into its corresponding
## invariant, i.e., \Red(L)
##
## L : tree list (fusion logic formula)
##
#proc reduce { L } {
#   global fresh
#   set root [lindex $L 0]
#   puts "reduce: L is $L"
#    if { [is_state $L] == 1 } {
#        return "{ Eqv r_1 {$L} }"
#    } else {
#        switch $root {
#            flchop { 
#                set E [lindex $L 1]
#                set F [lindex $L 2]
#                if { [is_state $F] == 1 } {
#                     puts "pass"
#                     set E_oper [lindex $E 0]
#                     set E_arg1 [lindex $E 1]
#                    if { $E_oper == "fetest" } {
#                        return " { Eqv r_1 { And {$F} {$E_arg1} } }"
#                    } else {
#                      if { $E_oper == "festep" } { 
#                          return " { Eqv r_1 { And {$E_arg1} {trnext {$F} } } }"#          
#                      } else {
#                        if { $E_oper == "feor" } { 
#                            set E_arg2 [lindex $E 2]
#                            set x "{ flor  { flchop {$E_arg1} {$F} } { flchop {$E_arg2} {$F} } }"
#                            set z [reduce [lindex $x 0]]
#                            return $z
#                        } else {
#                          if { $E_oper == "fechop" } { 
#                              set E_arg2 [lindex $E 2]
#                              set x "{ flchop {$E_arg1} { flchop {$E_arg2} {$F} } } "
#                            set z [reduce [lindex $x 0]]
#                            return $z
#                          } else {
#                            if { $E_oper == "fechopstar" } { 
#                              set y [c_fe $E_arg1 ]
#                              puts "y is $y"
#                              set x "{ flor {$F} { flchop $y r_q$fresh } }"
#                              puts "x is $x"
#                              set re "r_q$fresh"
#                              incr fresh
#                              set z [reduce [lindex $x 0]]
#                              set k [reduce_length [lindex $z 0] ]
#                              regsub -all $re $z {r_$k} z
#                              set z1 [subst $z]
#                              incr fresh -1
#                              return $z1
#                            } else { 
#                              puts "error in reduce F=W branch"
#                            }
#                          }
#                        }
#                      }
#                    }
#                } else {
#                  set x [reduce $F ]
#                  set k [reduce_length [lindex $x 0]]
#                  set y "{ flchop {$E} r_q$fresh }"
#                  set re "r_q$fresh"
#                  incr fresh
#                  set z [reduce [lindex $y 0]]
#                  set z1 [shift [lindex $z 0] $k]
#                  regsub -all $re $z1 {r_$k} z1
#                  set z2 [subst $z1] 
#                  incr fresh -1
#                  set z3 "{ And $x $z2 } "
#                  return $z3
#                }
#            }
#            flnot { 
#                     set x [reduce [lindex $L 1] ]
#                     set k [reduce_length [lindex $x 0]]
#                     set k1 [expr $k + 1]
#                     set z "{ And $x { Eqv r_$k1 {Not r_$k} } } "
#                     return $z
#            }
#            flor { 
#                    set x1 [reduce [lindex $L 1] ]
#                    set j [reduce_length [lindex $x1 0] ]
#                    set x2 [reduce [lindex $L 2] ]
#                    set y2 [shift [lindex $x2 0] $j]
#                    set k [reduce_length [lindex $x2 0]]
#                    set jk [expr $j + $k]
#                    set jk1 [expr $jk + 1]
#                    set z "{ And $x1 { And $y2 { Eqv r_$jk1 {Or r_$j r_$jk} } } }"
#                    return $z
#            }
#            default { puts "error in reduce switch" }
#        }
#    }
#}


#
# reverse_pl L
# time reversal of left fusion logic formula L\itlRev
# 
# input L: tree list (left fusion logic formula)
# output : right fusion logic formula L\itlRev
proc reverse_pl { L } {

   set root [lindex $L 0]
   # puts "reverse_pl: L is $L"

     switch $root {
         true { 
             set z "true"
             # puts "reverse_pl: true z is $z"
             return $z
         }
         false { 
             set z "false"
             # puts "reverse_pl: false z is $z"
             return $z
         }
         plchop { 
                  set E2 [lindex $L 2]
                  # puts $E2
                  set E1 [lindex $L 1]
                  # puts $E1
                  set z "{flchop [reverse_fe $E2] [reverse_pl $E1]}"
                  # puts "reverse_pl: plchop z is $z"
                  return $z
         }
         plfin { set Y [lindex $L 1]
                 set z "$Y"
                 # puts "reverse_pl: plfin z is $z"
                 return $z
         }
         plnot {
                 set x [reverse_pl [lindex $L 1]]
                 set z "{flnot $x}"
                 # puts "reverse_pl: plnot z is $z"
                 return $z
         }
         plor {
                 set x1 [reverse_pl [lindex $L 1] ]
                 set x2 [reverse_pl [lindex $L 2] ]
                 set z "{flor $x1 $x2}"
                 # puts "reverse_pl: plor z is $z"
                 return $z
         }
         pland {
                set x1 [reverse_pl [lindex $L 1] ]
                set x2 [reverse_pl [lindex $L 2] ]
                set z  "{fland $x1 $x2}"
                # puts "reverse_pl: pland z is $z"
                return $z
         }
         default { puts "error in reverse_pl switch" }
     }
}

#
# reverse_fe E
# time reversal of fusion expression E\itlRev
# 
# input L: tree list (fusion exxpression)
# output : fusion expression E\itlRev
proc reverse_fe { E } {

   set root [lindex $E 0]
   # puts "reverse_fe: E is $E"

     switch $root {
         fetest { 
             set Y [lindex $E 1]
             set z "{fetest {$Y}}"
             # puts "reverse_fe: fetest z is $z"
             return $z
         }
         festep {
             set Y [lindex $E 1]
             set z "{festep [reverse_tr $Y]}"
             # puts "reverse_fe: festep z is $z"
             return $z
         }
         fechop { 
                  set E1 [lindex $E 1]
                  set E2 [lindex $E 2]
                  set z "{fechop [reverse_fe $E2] [reverse_fe $E1]}"
                  # puts "reverse_fe: fechop z is $z"
                  return $z
         }
         fechopstar { 
                 set Y [lindex $E 1]
                 set z "{fechopstar [reverse_fe $Y]}"
                 # puts "reverse_fe: fechopstar z is $z"
                 return $z
         }
         feor {
                 set E1 [reverse_fe [lindex $E 1] ]
                 set E2 [reverse_fe [lindex $E 2] ]
                 set z "{feor $E1 $E2}"
                 # puts "reverse_fe: feor z is $z"
                 return $z
         }
         default { puts "error in reverse_fe switch" }
     }
}

#
# reverse_tr T
# time reversal of transition expression T\itlRev
# 
# input L: tree list (transition exxpression)
# output : transition expression T\itlRev
proc reverse_tr { T } {

   set root [lindex $T 0]
   # puts "reverse_tr: T is $T"
 
  switch -regexp -- $root {
      {true|false}     {
                        set z $root
                        # puts "reverse_tr: true-false z is $z"
                        return $z
      }
      {trand|tror}      {
                    set oper [lindex $T 0]
                    set x1 [reverse_tr [lindex $T 1] ]
                    set y1 [reverse_tr [lindex $T 2] ]
                    set z "{$oper $x1 $y1}"
                    # puts "reverse_tr: and-or z is $z"
                    return $z
      }
      {trnot}         {
                    set oper [lindex $T 0]
                    set x1 [reverse_tr [lindex $T 1] ]
                    set z "{$oper $x1}"
                    # puts "reverse_tr: not z is $z"
                    return $z
      }
      {trnext}        {
                     set x1 [lindex $T 1]
                     set z "$x1"
                     # puts "reverse_tr: next z is $z"
                     return $z
      } 
      {[A-Z]+[_A-Z]*}   { 
                   set z "{trnext $root}"
                   # puts "reverse_tr: var z is $z"
                   return $z
      }
      default { 
            puts "error in reverse_tr switch"    
      }
   }
}


#
# reduce_state k L
# reduction function v2 \Red'_k(L)
# 
# k: 0,1
# L: tree list (fusion logic formula)
#
proc reduce_state { k L } {

global depindex

   set root [lindex $L 0]
   #puts "reduce_state: L is $L"
   #puts "reduce_state: k is $k"

   if { [is_state $L] == 1 } { 
     #puts "reduce_state: is state"
     return "{$L}"
   } else {
     switch $root {
         flchop { 
                  set E [lindex $L 1]
                  set Y [lindex $L 2]
                  set E_oper [lindex $E 0]
                  set E_arg1 [lindex $E 1]
                  if { $E_oper == "fetest" } { 
                   set y1 [reduce_state $k $Y]
                   set z "{And {$E_arg1} $y1 }"
                   #puts "reduce_state: fetest z is $z"
                   return $z
                  } else {
                    if { $E_oper == "festep" } { 
                     if { $k == 0 } { 
                      return "r_$depindex($L)"
                     } else {
                       if { $k == 1 } { 
                        set y1 [reduce_state 0 $Y]
                        set z "{And {$E_arg1} {trnext $y1}}"
                        #puts "reduce_state: festep z is $z"
                        return $z
                       }
                     }
                    } else {
                      if { $E_oper == "feand" } {
                       set E_arg2 [lindex $E 2]
                       set y1 "{flchop {$E_arg2} {$Y}}"  
                       set y2 [reduce_state $k [lindex $y1 0]]
                       set z "{And {$E_arg1} $y2}"
                       #puts "reduce_state: feand z is $z"
                       return $z
                      } else {
                        if { $E_oper == "feor" } {
                         set E_arg2 [lindex $E 2]
                         set x "{flor {flchop {$E_arg1} {$Y}} {flchop {$E_arg2} {$Y}}}"
                         set z [reduce_state $k [lindex $x 0]]
                         #puts "reduce_state: feor z is $z"
                         return $z
                        } else { 
                          if { $E_oper == "fechop" } {
                           set E_arg2 [lindex $E 2]
                           set x "{flchop {$E_arg1} {flchop {$E_arg2} {$Y}}}"
                           set z [reduce_state $k [lindex $x 0]]
                           #puts "reduce_state: fechop z is $z"
                           return $z
                          } else {
                            if { $E_oper == "fechopstar" } {
                             set z "r_$depindex($L)"
                             #puts "reduce_state: fechopstar z is $z"
                             return $z
                            } else {
                              puts "error in reduce_state fusion expression branch"
                            }
                          }
                        }
                      }
                    }
                  }
         }
         flnot {
                 set x [reduce_state $k [lindex $L 1]]
                 set z "{Not $x}"
                 #puts "reduce_state: flnot z is $z"
                 return $z
         }
         flor {
                 set x1 [reduce_state $k [lindex $L 1] ]
                 set x2 [reduce_state $k [lindex $L 2] ]
                 set z "{Or $x1 $x2}"
                 #puts "reduce_state: flor z is $z"
                 return $z
         }
         fland {
                set x1 [reduce_state $k [lindex $L 1] ]
                set x2 [reduce_state $k [lindex $L 2] ]
                set z "{And $x1 $x2}"
                #set y1 "{flnot {flor {flnot { [lindex $L 1] }} {flnot { [lindex $L 2] }}}}"  
                #set z [reduce_state $k [lindex $y1 0]]
                #puts "reduce_state: fland z is $z"
                return $z
         }
         default { puts "error in reduce_state switch" }
     }
   }
}

#
# reduce_transition k L
# reduction function v2 \Red_k(L)
#
# k: 0,1
# L: tree list (fusion logic formula)
#
proc reduce_transition { k L } {
global depindex
global freshx
   set root [lindex $L 0]
   #puts "reduce_transition: L is $L"
   #puts "reduce_transition: k is $k"
   if { [is_state $L] == 1 } { 
      #puts "reduce_transition: is state"
      return "true"
   } else {
      switch $root {
          flchop {
                   set E [lindex $L 1]
                   set Y [lindex $L 2]
                   set E_oper [lindex $E 0]
                   set E_arg1 [lindex $E 1]
                   if { $E_oper == "fetest" } { 
                      set z [reduce_transition $k $Y]
                      #puts "reduce_transition: fetest z is $z"
                      return $z
                   } else { 
                     if { $E_oper == "festep" } { 
                      if { $k == 0 } { 
                        #set tmp [array names depindex -exact $L]
                        #puts "tmp is $tmp"
                        if {[array names depindex -exact $L] == "" } { 
                         #puts "assigning ($L) to $freshx"
                         set depindex($L) $freshx
                         incr freshx
                        } 
                        #else {
                        #  puts "reuse depindex ( )"
                        #}
                        set x2 [reduce_transition 0 $Y]
                        #puts "x2 is $x2"
                        set x1 [reduce_state 0 $Y]
                        #puts "x1 is $x1"
                        set z "{And $x2 {Eqv r_$depindex($L) {And {$E_arg1} {trnext $x1}}}}"
                        #puts "reduce_transition: festep z is $z"
                        return $z
                      } else { 
                        if { $k == 1 } { 
                         set z [reduce_transition 0 $Y]
                         #puts "reduce_transition: festep z is $z"
                         return $z
                        } else { 
                         puts " error in festep of reduce_transition "
                        }
                      }
                     } else {
                       if { $E_oper == "feand" } { 
                        set E_arg2 [lindex $E 2]
                        set y1 "{flchop {$E_arg2} {$Y}}" 
                        set z [reduce_transition $k [lindex $y1 0]]
                        return $z
                       } else {
                         if { $E_oper == "feor" } { 
                          set E_arg2 [lindex $E 2]
                          set x "{flor {flchop {$E_arg1} {$Y}} {flchop {$E_arg2} {$Y}}}"
                          #puts "reduce_transition: feor z is $z"
                          set z [reduce_transition $k [lindex $x 0]]
                          return $z
                         } else {
                           if { $E_oper == "fechop" } { 
                            set E_arg2 [lindex $E 2]
                            set x "{flchop {$E_arg1} {flchop {$E_arg2} {$Y}}}"
                            set z [reduce_transition $k [lindex $x 0]]
                            #puts "reduce_transition: fechop z is $z"
                            return $z
                           } else {
                             if { $E_oper == "fechopstar" } {
                              #set tmp [array names depindex -exact $L]
                              #puts "tmp is $tmp"
                              if {[array names depindex -exact $L] == "" } { 
                               #puts "assigning ($L) to $freshx"
                               set depindex($L) $freshx
                               incr freshx
                              } 
                              #else {
                              # puts "reuse depindex ( )"
                              #}
                              set y1 [c_fe $E_arg1 ]
                              set x1 "{flor {$Y} {flchop $y1 r_$depindex($L)}}"
                              set y3 [reduce_transition 1 [lindex $x1 0]]
                              set y2 [reduce_state 1 [lindex $x1 0]]
                              set z1 "{Eqv r_$depindex($L) $y2}"
                              set z "{And $z1 $y3} "
                              #puts "reduce_transition: fechopstar z is $z"
                              return $z
                             } else {
                               puts "error in reduce_transition fusion expression branch"
                             }
                           }
                         }
                       }
                     }
                   }
          }
          flnot { 
                   set z [reduce_transition $k [lindex $L 1] ]
                   #puts "reduce_transition: flnot z is $z"
                   return $z
                }
          flor {
                   set x1 [reduce_transition $k [lindex $L 1] ]
                   set x2 [reduce_transition $k [lindex $L 2] ]
                   set z "{And $x1 $x2}"
                   #puts "reduce_transition: flor z is $z"
                   return $z
               }
          fland {
                   set x1 [reduce_transition $k [lindex $L 1] ]
                   set x2 [reduce_transition $k [lindex $L 2] ]
                   set z "{And $x1 $x2}"
                   #set y1 "{flnot {flor {flnot { [lindex $L 1] }} {flnot { [lindex $L 2] }}}}" 
                   #set z [reduce_transition $k [lindex $y1 0]]
                   return $z
          }
          default { puts "error in reduce_transition switch" }
      }
   }
}


#
# next_sub_replace L
# replace all variables in L by their primed version
#
# L: tree list (state formula from next_sub)
#
proc next_sub_replace { L } {
  set root [lindex $L 0]

  #puts "next_sub_replace: L is $L"
 
  switch -regexp -- $root {
      {And|Or}      {
                    set oper [lindex $L 0]
                    set x1 [next_sub_replace [lindex $L 1] ]
                    set y1 [next_sub_replace [lindex $L 2] ]
                    set z "{$oper $x1 $y1}"
                    return $z
      }
      {Not}         {
                    set oper [lindex $L 0]
                    set x1 [next_sub_replace [lindex $L 1] ]
                    set z "{$oper $x1}"
                    return $z
      }
      {r_[0-9]+} { 
                    set j [string range $root 2 end]
                    return " pr_$j "
      }
      {[A-Z]+[_A-Z]*}   { 
                   return " p$root "
      }
      default { 
                if {[llength $L] == 1 } {
                  return "{$L}"
                } else {
                   if {[llength $L] == 2 } {
                     set oper [lindex $L 0]
                     set x [next_sub_replace [lindex $L 1] ]
                     return "{$oper $x}"
                   } else {
                     if {[llength $L] == 3 } {
                       set oper [lindex $L 0]
                       set x [next_sub_replace [lindex $L 1] ]
                       set y [next_sub_replace [lindex $L 2] ]
                       return "{$oper $x $y}"
                     } else {
                       puts "error in next_sub_replace"
                     }
                   }
                }
      }
  }
}

# 
# next_sub L
#  find formula in scope of a next
# 
# L : tree list (fusion logic formula)
#
proc next_sub { L } {
set root [lindex $L 0]
  
 
  switch  $root {
      trnext { 
                set x [next_sub_replace [lindex $L 1] ]
                return $x
      }
      default { 
                if {[llength $L] == 1 } {
                  return "{$L}"
                } else {
                   if {[llength $L] == 2 } {
                     set oper [lindex $L 0]
                     set x [next_sub [lindex $L 1] ]
                     return "{$oper $x}"
                   } else {
                     if {[llength $L] == 3 } {
                       set oper [lindex $L 0]
                       set x [next_sub [lindex $L 1] ]
                       set y [next_sub [lindex $L 2] ]
                       return "{$oper $x $y}"
                     } else {
                       puts "error in next_sub"
                     }
                   }
                }
      }
  }           
}

#
# next_sub_false L
#  replace all variables in L within scope of a next by false
#
# L: tree list (fusion logic formula)
#
proc next_sub_false { L } {
set root [lindex $L 0]
   
  switch  $root {
      trnext { 
                return " false "
      }
      default { 
                if {[llength $L] == 1 } {
                  return "{$L}"
                } else {
                   if {[llength $L] == 2 } {
                     set oper [lindex $L 0]
                     set x [next_sub_false [lindex $L 1] ]
                     return "{$oper $x}"
                   } else {
                     if {[llength $L] == 3 } {
                       set oper [lindex $L 0]
                       set x [next_sub_false [lindex $L 1] ]
                       set y [next_sub_false [lindex $L 2] ]
                       return "{$oper $x $y}"
                     } else {
                       puts "error in next_sub_false"
                     }
                   }
                }
      }
  }            
}

# tranform reduced tree list to BDD routines
#

#
# flatten args
#  sanitize the tree list args
#
# args: tree list
#
proc flatten args {
    regsub -all {[{}]} $args {} tmp;    # Ditch braces
    regsub -all { +} $tmp { } tmp;      # Squeeze spaces
    set fl [string trim $tmp];          # Chop leading/trailing white space
    return $fl
}

#
# is_var L
#  check whether L is a variable
# 
# L: tree list (fusion logic formula)
#
proc is_var { L } {
  set root [lindex $L 0]
  #puts "is_var: root is $root" 
  switch -regexp -- $root {
    {r_[0-9]+} { 
                   return 1
    }
    {pr_[0-9]+} { 
                   return 1
    }
    {p[A-Z]+[_A-Z]*} { 
                   return 1
    }
    {r[A-Z]+[_A-Z]*} { 
                   return 1
    }
    
    default { 
                return 0
      }
  }

}
 

#
# extract_var L
#  extract the variables used in formula L
#
# L: tree list (fusion logic formula)
#
proc extract_var { L } {
  set root [lindex $L 0]
  #puts "extract_var: root is $root" 
  switch -regexp -- $root {
      {Eqv|And|Or|Not} { 
                     if {[llength $L] == 1 } {
                     return ""
                     } else {
                      if {[llength $L] == 2 } {
                        set oper [lindex $L 0]
                        set x [extract_var [lindex $L 1] ]
                        return $x
                      } else {
                        if {[llength $L] == 3 } {
                          set oper [lindex $L 0]
                          set x [extract_var [lindex $L 1] ]
                          lappend z $x
                          set y [extract_var [lindex $L 2] ]
                          lappend z $y
                          return $z
                        } else {
                          puts "error in extract_var"
                        }
                      }
                     }
      }
      {r_[0-9]+} { 
                   return $root
      }
      {[A-Z]+[_A-Z]*}   { 
                   return $root
      }
      default { 
                if {[llength $L] == 1 } {
                  return ""
                } else {
                   if {[llength $L] == 2 } {
                     set oper [lindex $L 0]
                     set x [extract_var [lindex $L 1] ]
                     return $x
                   } else {
                     if {[llength $L] == 3 } {
                       set oper [lindex $L 0]
                       set x [extract_var [lindex $L 1] ]
                       lappend z $x
                       set y [extract_var [lindex $L 2] ]
                       lappend z $y
                       return $z
                     } else {
                       puts "error in extract_var"
                     }
                   }
                }
      }
  }
}

#
# normalise L
#  replace non-primed independent variables A by pA in L
#
# L: tree list (fusion logic formula)
# 
proc normalise { L } {
  set root [lindex $L 0]
  set lroot [llength $L]
  
  switch $lroot {
      1   { 
          set x [lindex $L 0]
          if {[string index $x 0] == "p" } {
             return $x
          } else {
             regsub -all {([A-Z]+[_A-Z]*)} $x {r\1} x
             return $x
          }
      }
      2   { 
          set oper [lindex $L 0]
          set x [normalise [lindex $L 1]]
          return [list $oper $x]
      }
      3   { 
          set oper [lindex $L 0]
          set x [normalise [lindex $L 1]]
          set y [normalise [lindex $L 2]]
          return [list $oper $x $y]
      }
      default { 
                return "error in normalise"
      }
  }
}

#
# tree2bdd_init gamma
#  replace dependent variables in gamma by their bdd equivalent 
#
#  gamma: tree list (init of reduced fusion logic formula)
proc tree2bdd_init { gamma } {
global a
global b
variable indep
    
set gammaa [regsub -all {(r_)([0-9]+)} $gamma {[lindex $a [expr {\2 -1}]]}]
return $gammaa
}

#
# tree2bdd gamma
#  transform gamma into a bdd expression
#
# gamma: tree list (reduced fusion logic formula)
#
proc tree2bdd { gamma } {
global a
global b
variable indep

set gammaa [regsub -all {\{true\}} $gamma {[One]}]
set gammaa [regsub -all {\{false\}} $gammaa {[Zero]}]
set gammaa [regsub -all {\{flnot true\}} $gammaa {[Zero]}]
set gammaa [regsub -all {\{} $gammaa {[}]
set gammaa [regsub -all {\}} $gammaa {]}]
set gammaa [regsub -all {(fl|st|tr)or} $gammaa {Or}]
set gammaa [regsub -all {(fl|st|tr)and} $gammaa {And}]
set gammaa [regsub -all {(fl|st|tr)not} $gammaa {Not}] 
set gammaa [regsub -all {true} $gammaa {[One]}]
set gammaa [regsub -all {false} $gammaa {[Zero]}]
set gammaa [regsub -all {(pr_)([0-9]+)} $gammaa {[lindex $b [expr {\2 -1}]]}]
set gammaa [regsub -all {(r_)([0-9]+)} $gammaa {[lindex $a [expr {\2 -1}]]}]
set gammaa [regsub -all {p([A-Z]+[_A-Z]*)} $gammaa {[lindex $b $indep(\1)]}]
set gammaa [regsub -all {r([A-Z]+[_A-Z]*)} $gammaa {[lindex $a $indep(\1)]}]

return $gammaa

}

#
# input_to_trace inputlist
#  convert input to enforcer to a bdd expression
#
# inputList: tree list (input to enforcer)
#
proc input_to_trace { inputList } {
global a
global b
variable indep

set trace_a [regsub -all {\{} $inputList {[}]
set trace_a [regsub -all {\}} $trace_a {]}]
set trace_a [regsub -all {([A-Z]+[_A-Z]*)} $trace_a {[lindex $a $indep(\1)]}]
set trace_a [regsub -all {not} $trace_a {Not}]
#set trace_a [regsub -all {tand} $trace_a {And}]
set trace_a [regsub -all {and} $trace_a {And}]
set trace_a [regsub -all {true} $trace_a {[One]}]
return $trace_a
}


# GUI routines
#


proc openfile {} {
  global current

#    variable TEMPURA
#    global term_spawn_id
    set types {
    { {TCL File}           {.tcl}}
    }
    set tmp $current
    set current \
        [tk_getOpenFile -defaultextension ".tcl" -filetypes $types ]
    if {$current != ""} {
      set ttt [file normalize $current]
      wm title .top18 "FLCHECK: $ttt"
    } else {
      set current $tmp
    }

}

proc load_file { lfile } {
global current

set curdir [file dir $current]

set curfile [file join $curdir $lfile]

 if {![file exists $curfile]} {
   insert "File $curfile not found!"

 } else {
   set ttt [file normalize $curfile]
   insert "Loading $ttt"
   source $curfile
 }  

}

proc check_file { cfile } {
  global current

 if {![file exists $cfile]} {
   set ttt [file normalize $cfile]
   insert "File $ttt not found!"
   set current ""
 } else {
   set current $cfile
 }  

}

proc Window {args} {
#global vTcl
    set cmd [lindex $args 0]
    set name [lindex $args 1]
    set newname [lindex $args 2]
    set rest [lrange $args 3 end]
    if {$name == "" || $cmd == ""} {return}
    if {$newname == ""} {
        set newname $name
    }
    set exists [winfo exists $newname]
    switch $cmd {
        show {
            if {$exists == "1" && $name != "."} {wm deiconify $name; return}
            if {[info procs flcheckWindow(pre)$name] != ""} {
                eval "flcheckWindow(pre)$name $newname $rest"
            }
            if {[info procs flcheckWindow$name] != ""} {
                eval "flcheckWindow$name $newname $rest"
            }
            if {[info procs flcheckWindow(post)$name] != ""} {
                eval "flcheckWindow(post)$name $newname $rest"
            }
        }
        hide    { if {$exists} {wm withdraw $newname; return} }
        iconify { if {$exists} {wm iconify $newname; return} }
        destroy { if {$exists} {destroy $newname; return} }
    }
}

proc flcheckWindow. {base} {
    if {$base == ""} {
        set base .
    }
    ###################
    # CREATING WIDGETS
    ###################
#    wm focusmodel $base passive
    wm geometry $base 122x26+0+0
    wm maxsize $base 1137 870
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm withdraw $base
    wm title $base "BDD"
    ###################
    # SETTING GEOMETRY
    ###################
}


proc flcheckWindow.top18 {base} {
    global out help term enter_text entry_widget notebook_widget
    global bddfont tback back current
      
    if {$base == ""} {
        set base .top18
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    ###################
    # CREATING WIDGETS
    ###################
    toplevel $base -class Toplevel  
#    wm focusmodel $base passive
    wm geometry $base 650x495+1+1
    wm maxsize $base 1137 870
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base "FLCHECK:$current"

    ttk::frame $base.f

   
    menu $base.f.fmenu

    $base.f.fmenu add cascade -label "File" \
       -font $bddfont\
       -underline 0 -menu $base.f.fmenu.entry
    $base configure -menu $base.f.fmenu

    menu $base.f.fmenu.entry -relief raised  \
        -tearoff 0


    $base.f.fmenu.entry add command \
       -font $bddfont\
       -label Open -command {::bdd_tcl::openfile}

    
#    $base.f.fmenu.entry add command \
#        -font $bddfont\
#        -label Reload -command {source  fusion_logic_derived.tcl}

    $base.f.fmenu.entry add command \
        -font $bddfont\
        -label Exit -command {exit}

    ttk::frame $base.f1  \
        -borderwidth 1 -height 405 -relief raised \
        -width 169 

    ttk::scrollbar $base.f1.01 \
        -command [list $base.f1.03 xview] -orient horiz  
    ttk::scrollbar $base.f1.02  \
        -command [list $base.f1.03 yview] -orient vert 
    text $base.f1.03 -background $tback \
        -font $bddfont\
       -cursor xterm -borderwidth 1 \
       -highlightthickness 0\
       -wrap none -state disabled\
       -height 7 \
       -insertbackground black \
       -selectbackground black \
       -selectforeground $back\
       -xscrollcommand [list $base.f1.01 set]\
       -yscrollcommand [list $base.f1.02 set]
        

    set entry_widget $base.f1.04

    history::for ttk::entry $entry_widget -font $bddfont -textvar enter_text
    
    #ttk::button $base.f1.exit   -text "Exit"  -command "exit"

    set term $base.f1.03

    

    ttk::notebook $base.f1.n

    set notebook_widget $base.f1.n

    ttk::frame $base.f1.n.f1
    set frm $base.f1.n.f1

    $base.f1.n add $frm -text "Output"

    ttk::scrollbar $frm.04\
        -command [list $frm.06 xview] -orient horiz
    ttk::scrollbar $frm.05 \
        -command [list $frm.06 yview] -orient vert
    text $frm.06 -background $tback\
       -font $bddfont\
       -cursor xterm -borderwidth 1 \
       -highlightthickness 0\
       -wrap none -state disabled\
       -height 17 \
       -insertbackground black \
       -selectbackground black \
       -selectforeground $back\
       -xscrollcommand [list $frm.04 set]\
       -yscrollcommand [list $frm.05 set]

    
    set out $frm.06  

    ttk::frame $base.f1.n.f2
    set frmh $base.f1.n.f2

    $base.f1.n add $frmh -text "Help"

    ttk::scrollbar $frmh.04\
        -command [list $frmh.06 xview] -orient horiz
    ttk::scrollbar $frmh.05 \
        -command [list $frmh.06 yview] -orient vert
    text $frmh.06 -background $tback\
       -font $bddfont\
       -cursor xterm -borderwidth 1 \
       -highlightthickness 0\
       -wrap none -state disabled\
       -height 17 \
       -insertbackground black \
       -selectbackground black \
       -selectforeground $back\
       -xscrollcommand [list $frmh.04 set]\
       -yscrollcommand [list $frmh.05 set]


    set help $frmh.06 
    
    pack $base.f \
        -in $base -anchor center -expand 0 -fill x -side top
    
    pack $base.f1 \
        -in $base -anchor center -expand 1 -fill both -side top

    

    grid columnconf $base.f1 0 -weight 1
     grid rowconf $base.f1 0 -weight 1
   #  grid $base.f1.exit \
   #     -in $base.f1 -column 0 -row 0 -columnspan 2 -rowspan 1 \
   #     -sticky nsew
     grid $base.f1.n \
       -in $base.f1 -row 0 -column 0 -columnspan 2 -rowspan 1 -sticky nsew
     grid $base.f1.01 \
       -in $base.f1 -row 2 -column 0 -columnspan 1 -rowspan 1 -sticky nsew
     grid $base.f1.02 \
       -in $base.f1 -row 1 -column 1 -columnspan 1 -rowspan 1 -sticky nsew
     grid $base.f1.03 \
       -in $base.f1 -row 1 -column 0 -columnspan 1 -rowspan 1 -sticky nsew
     grid $base.f1.04 \
        -in $base.f1 -column 0 -row 3 -columnspan 2 -rowspan 1 \
        -sticky nesw
     
     

    grid columnconf $frm 0 -weight 1
    grid rowconf $frm 0 -weight 1
    grid $frm.04 \
        -in $frm -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky ew 
    grid $frm.05 \
        -in $frm -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
    grid $frm.06 \
        -in $frm -column 0 -row 0 -columnspan 1 -rowspan 1 \
        -sticky nesw

    grid columnconf $frmh 0 -weight 1
    grid rowconf $frmh 0 -weight 1
    grid $frmh.04 \
        -in $frmh -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky ew 
    grid $frmh.05 \
        -in $frmh -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
    grid $frmh.06 \
        -in $frmh -column 0 -row 0 -columnspan 1 -rowspan 1 \
        -sticky nesw
     

}

proc insert {txt {tag normal} } {
  global out nodisplay
  if { $nodisplay == 0 } { 
    $out configure -state normal
    $out insert end $txt $tag
    $out see end
    $out insert end "\n"
    $out see end
    $out configure -state disabled
    update idletasks
  } else {
    puts $txt
  }
 
}

proc insert_help {txt {tag normal} } {
  global help  

    $help configure -state normal
    $help insert end $txt $tag
    $help see end
    $help insert end "\n"
    $help see end
    $help configure -state disabled
   
}

proc help_begin {} {
  global help  

    $help see 1.0
   
}

proc insert_input {txt {tag normal} } {
  global term nodisplay
  if { $nodisplay == 0 } { 
    $term configure -state normal
    $term insert end $txt $tag
    $term see end
    $term insert end "\n"
    $term see end
    $term configure -state disabled
    update idletasks
  } else {
   puts $txt
  }
}

#proc insert1 {c} {
#    global cmd
#    global point
#
#    set cmd [linsert $cmd $point $c]
#    incr point
#}

#proc backspace {} {
#    global cmd
#    global point
#
#    if {0 < $point} {
#        incr point -1
#        set cmd [lreplace $cmd $point $point]
#    }
#
#    return
#}

#proc right {} {
#    global cmd
#    global point
#
#    if {$point < [llength $cmd]} {
#        incr point 1
#    }
#
#    return
#}

#proc left {} {
#    global point
#
#    if {0 < $point} {
#        incr point -1
#    }
#
#    return
#}

#proc up {} {
#    global cmd
#    global point
#    global event
#    global saveCmd
#    global term
#
    # If we're leaving event 1 (the new command) save it.
#
#    if {$event == 1} {
#        set saveCmd [join $cmd ""]
#    }
#
#    # Erase this command and get the previous event.
#
#    control-u
#    $term insert end "\n"
#    if {[expr {1 - [history keep]}] < $event} {
#        incr event -1
#    }
#    set cmd [history event $event]
#    $term insert insert $cmd
#    $term see insert
#    set cmd [split $cmd {}]
#    set point [llength $cmd]
#
#    return
#}

#proc control-u {} {
#    global cmd
#    global point
#    global term
#
#    while {$point > 0} {
#         $term mark set insert insert-1c
#        incr point -1
#    }
#    $term delete insert end
#    $term insert end "\n"
#    $term see end
#    set cmd {}
#
#    return
#}

#proc down {} {
#    global cmd
#    global point
#    global event
#    global saveCmd
#    global term
#
#    # Erase this command and get the next event.
#
#    if {$event == 1} {
#        return
#    }
#    control-u
#    $term insert end "\n"
#    incr event
#    if {$event == 1} {
#        set cmd $saveCmd
#    } else {
#        set cmd [history event $event]
#    }
#
#    $term insert insert $cmd
#    $term see insert
#    set cmd [split $cmd {}]
#    set point [llength $cmd]
#
#}

proc remove_tabs {base} {
  global tab
  #puts "tab: $tab"
  for {set i 3} {$i<$tab} {incr i} {
   #puts "remove tab $i"
   $base.f1.n forget $base.f1.n.f$i
  }
  set tab 3
}

proc new_tab {base fnr {tabtext "BDD"}} {
   global c 
   catch { ttk::frame $base.f1.n.f$fnr }

    set frm $base.f1.n.f$fnr
    if {$tabtext == "BDD"} {
        set tabtext "$tabtext [expr $fnr -5]"
    }
    $base.f1.n add $frm -text $tabtext 

    catch { ttk::scrollbar $frm.04\
        -command [list $frm.06 xview] -orient horiz }
    catch { ttk::scrollbar $frm.05 \
        -command [list $frm.06 yview] -orient vert }
    catch { canvas $frm.06 -width 300 -height 200 -background white\
       -highlightthickness 0\
       -xscrollcommand [list $frm.04 set]\
       -yscrollcommand [list $frm.05 set] }
    
    set c $frm.06

    
     grid $frm.04 \
        -in $frm -column 0 -row 1 -columnspan 3 -rowspan 1 -sticky ew 
     grid $frm.05 \
        -in $frm -column 3 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
     grid $frm.06 \
        -in $frm -column 0 -row 0 -columnspan 3 -rowspan 1 \
        -sticky nesw 
     grid columnconf $frm 0 -weight 1 
     grid columnconf $frm 1 -weight 1 
     grid columnconf $frm 2 -weight 1 
     grid rowconf $frm 0 -weight 1 
    catch { ttk::button $frm.zoomin   -text "Zoom in"  -command "::bdd_tcl::zoom $c 1.25" }
    catch { ttk::button $frm.zoomout  -text "Zoom out" -command "::bdd_tcl::zoom $c 0.8" }
     grid $frm.zoomin $frm.zoomout -sticky ew
    # Set up event bindings for canvas:
    bind $c <3> "::bdd_tcl::zoomMark $c %x %y"
    bind $c <B3-Motion> "::bdd_tcl::zoomStroke $c %x %y"
    bind $c <ButtonRelease-3> "::bdd_tcl::zoomArea $c %x %y"


}

#
# zoom routines from http://wiki.tcl.tk/4844
#

#--------------------------------------------------------
#
#  zoomMark
#
#  Mark the first (x,y) coordinate for zooming.
#
#--------------------------------------------------------
proc zoomMark {c x y} {
    global zoomArea
    set zoomArea(x0) [$c canvasx $x]
    set zoomArea(y0) [$c canvasy $y]
    $c create rectangle $x $y $x $y -outline black -tag zoomArea
}

#--------------------------------------------------------
#
#  zoomStroke
#
#  Zoom in to the area selected by itemMark and
#  itemStroke.
#
#--------------------------------------------------------
proc zoomStroke {c x y} {
    global zoomArea
    set zoomArea(x1) [$c canvasx $x]
    set zoomArea(y1) [$c canvasy $y]
    $c coords zoomArea $zoomArea(x0) $zoomArea(y0) $zoomArea(x1) $zoomArea(y1)
}

#--------------------------------------------------------
#
#  zoomArea
#
#  Zoom in to the area selected by itemMark and
#  itemStroke.
#
#--------------------------------------------------------
proc zoomArea {c x y} {
    global zoomArea

    #--------------------------------------------------------
    #  Get the final coordinates.
    #  Remove area selection rectangle
    #--------------------------------------------------------
    set zoomArea(x1) [$c canvasx $x]
    set zoomArea(y1) [$c canvasy $y]
    $c delete zoomArea

    #--------------------------------------------------------
    #  Check for zero-size area
    #--------------------------------------------------------
    if {($zoomArea(x0)==$zoomArea(x1)) || ($zoomArea(y0)==$zoomArea(y1))} {
        return
    }

    #--------------------------------------------------------
    #  Determine size and center of selected area
    #--------------------------------------------------------
    set areaxlength [expr {abs($zoomArea(x1)-$zoomArea(x0))}]
    set areaylength [expr {abs($zoomArea(y1)-$zoomArea(y0))}]
    set xcenter [expr {($zoomArea(x0)+$zoomArea(x1))/2.0}]
    set ycenter [expr {($zoomArea(y0)+$zoomArea(y1))/2.0}]

    #--------------------------------------------------------
    #  Determine size of current window view
    #  Note that canvas scaling always changes the coordinates
    #  into pixel coordinates, so the size of the current
    #  viewport is always the canvas size in pixels.
    #  Since the canvas may have been resized, ask the
    #  window manager for the canvas dimensions.
    #--------------------------------------------------------
    set winxlength [winfo width $c]
    set winylength [winfo height $c]

    #--------------------------------------------------------
    #  Calculate scale factors, and choose smaller
    #--------------------------------------------------------
    set xscale [expr {$winxlength/$areaxlength}]
    set yscale [expr {$winylength/$areaylength}]
    if { $xscale > $yscale } {
        set factor $yscale
    } else {
        set factor $xscale
    }

    #--------------------------------------------------------
    #  Perform zoom operation
    #--------------------------------------------------------
    zoom $c $factor $xcenter $ycenter $winxlength $winylength
}

#--------------------------------------------------------
#
#  zoom
#
#  Zoom the canvas view, based on scale factor
#  and centerpoint and size of new viewport.
#  If the center point is not provided, zoom
#  in/out on the current window center point.
#
#  This procedure uses the canvas scale function to
#  change coordinates of all objects in the canvas.
#
#--------------------------------------------------------
proc zoom { canvas factor \
    {xcenter ""} {ycenter ""} \
    {winxlength ""} {winylength ""} } {

    #--------------------------------------------------------
    #  If (xcenter,ycenter) were not supplied,
    #  get the canvas coordinates of the center
    #  of the current view.  Note that canvas
    #  size may have changed, so ask the window
    #  manager for its size
    #--------------------------------------------------------
    if { [string equal $xcenter ""] } {
        set winxlength [winfo width $canvas]
        set winylength [winfo height $canvas]
        set xcenter [$canvas canvasx [expr {$winxlength/2.0}]]
        set ycenter [$canvas canvasy [expr {$winylength/2.0}]]
    }

    #--------------------------------------------------------
    #  Scale all objects in the canvas
    #  Adjust our viewport center point
    #--------------------------------------------------------
    $canvas scale all 0 0 $factor $factor
    set xcenter [expr {$xcenter * $factor}]
    set ycenter [expr {$ycenter * $factor}]

    #--------------------------------------------------------
    #  Get the size of all the items on the canvas.
    #
    #  This is *really easy* using
    #      $canvas bbox all
    #  but it is also wrong.  Non-scalable canvas
    #  items like text and windows now have a different
    #  relative size when compared to all the lines and
    #  rectangles that were uniformly scaled with the
    #  [$canvas scale] command.
    #
    #  It would be better to tag all scalable items,
    #  and make a single call to [bbox].
    #  Instead, we iterate through all canvas items and
    #  their coordinates to compute our own bbox.
    #--------------------------------------------------------
    set x0 1.0e30; set x1 -1.0e30 ;
    set y0 1.0e30; set y1 -1.0e30 ;
    foreach item [$canvas find all] {
        switch -exact [$canvas type $item] {
       	"arc" -
       	"line" -
       	"oval" -
       	"polygon" -
       	"rectangle" {
       	        set coords [$canvas coords $item]
       	        foreach {x y} $coords {
       		    if { $x < $x0 } {set x0 $x}
       		    if { $x > $x1 } {set x1 $x}
       		    if { $y < $y0 } {set y0 $y}
       		    if { $y > $y0 } {set y1 $y}
       	        }
       	     }
        }
    }

    #--------------------------------------------------------
    #  Now figure the size of the bounding box
    #--------------------------------------------------------
    set xlength [expr {$x1-$x0}]
    set ylength [expr {$y1-$y0}]

    #--------------------------------------------------------
    #  But ... if we set the scrollregion and xview/yview
    #  based on only the scalable items, then it is not
    #  possible to zoom in on one of the non-scalable items
    #  that is outside of the boundary of the scalable items.
    #
    #  So expand the [bbox] of scaled items until it is
    #  larger than [bbox all], but do so uniformly.
    #--------------------------------------------------------
    foreach {ax0 ay0 ax1 ay1} [$canvas bbox all] {break}

    while { ($ax0<$x0) || ($ay0<$y0) || ($ax1>$x1) || ($ay1>$y1) } {
        # triple the scalable area size
        set x0 [expr {$x0-$xlength}]
        set x1 [expr {$x1+$xlength}]
        set y0 [expr {$y0-$ylength}]
        set y1 [expr {$y1+$ylength}]
        set xlength [expr {$xlength*3.0}]
        set ylength [expr {$ylength*3.0}]
    }

    #--------------------------------------------------------
    #  Now that we've finally got a region defined with
    #  the proper aspect ratio (of only the scalable items)
    #  but large enough to include all items, we can compute
    #  the xview/yview fractions and set our new viewport
    #  correctly.
    #--------------------------------------------------------
    set newxleft [expr {($xcenter-$x0-($winxlength/2.0))/$xlength}]
    set newytop  [expr {($ycenter-$y0-($winylength/2.0))/$ylength}]
    $canvas configure -scrollregion [list $x0 $y0 $x1 $y1]
    $canvas xview moveto $newxleft
    $canvas yview moveto $newytop

    #--------------------------------------------------------
    #  Change the scroll region one last time, to fit the
    #  items on the canvas.
    #--------------------------------------------------------
    $canvas configure -scrollregion [$canvas bbox all]
}

#
# check inp
#  check validity/satisfaction/enforcement of formula inp
#
# inp: string (fusion logic formula plus type of check)
#
proc check { inp} {
  variable a
  variable b  
  variable indep
  global c tab fresh freshx depindex INPUT nodisplay current
  
     
    set fresh 0
    set freshx 1
    array unset depindex
    array unset indep
    
    
      if {$current != ""} {
        set ttt [file normalize $current]
        insert "Loading $ttt"
        source $current
      }
      insert "input = $inp"
    
    
      set inp1 [subst $inp]
      yy_scan_string $inp1
      set loc [yyparse]
    
      if {$loc != "error"} { 
      Init-data-manager
      set tl [lindex $loc 1]
      set check [lindex $loc 0]
      #puts "results: $loc"
     
      ##set tl1 [reduce [lindex $tl 0]]
      if {$check > 3} { 
           set tlreverse [reverse_pl [lindex $tl 0]]
           set tl1 [reduce_transition 0 [lindex $tlreverse 0]]
           set tlst [reduce_state 0 [lindex $tlreverse 0]]
      } else {
         set tl1 [reduce_transition 0 [lindex $tl 0]]    
         set tlst [reduce_state 0 [lindex $tl 0]]
      }
      #puts "tl1: $tl1"
      #puts "tlst: $tlst"
    
      #set k [reduce_length [lindex $tl1 0] ]
    
      set var_for1 [flatten [extract_var [lindex $tlst 0]]]    
      #puts "var_for1 is $var_for1"
      set var_for2 [flatten [extract_var [lindex $tl1 0]]]
      #puts "var_for2 is $var_for2"
      set var_for_3 [concat $var_for1 $var_for2]
      set var_for  [lsort -unique $var_for_3]
      set var_for  [lsort -dictionary $var_for]
      #puts "var_for is $var_for"
      set last_el  [lindex $var_for end]
      #puts "last_el is $last_el"
      if {$last_el == {} } {
         set k 0
      } else {
        if { [string index $last_el 0] == "r" } {
         set k [string range $last_el 2 end]
        } else {
          set k 0
        }
      }
      #puts "k is $k"
      #puts "var_for is $var_for"
    

      if {$k == 0} {
        set dep_var_for {} 
        set indep_var_for $var_for
      } else { 
      
        set k1 [expr $k-1]
        #puts "k1 is $k1"
    
        ##set gamma1 "r_$k"
        set dep_var_for [lrange $var_for end-$k1 end]
        set indep_var_for [lrange $var_for 0 end-$k]
      }
      #puts "dep_var_for is $dep_var_for"
      #puts "indep_var_for is $indep_var_for"
    
      #puts "gamma1 is before $tlst"
      set gamma1 [normalise [lindex $tlst 0]]
      if { [is_var $gamma1] == 1 } {
        set gamma1 $gamma1
      } else {
        set gamma1 "{$gamma1}"
      }
      #puts "gamma1 is after $gamma1" 
      set gamma2 [normalise [lindex [next_sub [lindex $tl1 0] ] 0] ]
      set gamma2 "{$gamma2}"
      
      set gamma3 [normalise [lindex [next_sub_false [lindex $tl1 0] ] 0] ]
      set gamma3 "{$gamma3}"

      insert "dependent variables : $dep_var_for"
     
      insert "independent variables : "
       set zz [split $indep_var_for " "]
       foreach zzz $zz {
         if { [string range $zzz 0 5] == "AUTHON" } {
           insert "  $zzz" red
         } else { 
           if { [string range $zzz 0 5] == "AUTHOP" } {
             insert "  $zzz" green
           } else {
             insert "  $zzz"
             }
         }
       }
    
      if {$nodisplay==0} {     
        drawtree $tl  "Input"
        if {$check>3} {
          drawtree $tlreverse "Reverse"
        }
        drawtree $tl1 "Reduce"
        drawtree $gamma1 "Gamma1"
        drawtree $gamma2 "Gamma2"
        drawtree $gamma3 "Gamma3"
        update idletasks
      }

      set i $k
      foreach x $indep_var_for {
          array set indep [list $x $i]
          incr i
      }

      set suffix_list [concat $dep_var_for $indep_var_for]

      #puts "suffix_list is $suffix_list"
     
      set t_init [time {init_my_basic_vars $suffix_list}]
      insert "init time: $t_init"
      update idletasks
    
      #insert "start gamma1 manipulations"
      #puts "gamma1 is $gamma1"
      set gamma1a [tree2bdd $gamma1]
      #puts "gamma1a: $gamma1a"
      #insert "start gamma2 manipulations"
      #puts "gamma2 is $gamma2"
      set gamma2a [tree2bdd $gamma2]
      #puts "gamma2a: $gamma2a"
      #insert "start gamma3 manipulations"
      #puts "gamma3 is $gamma3"
      set gamma3a [tree2bdd $gamma3]
      #puts "gamma3a: $gamma3a"

      set InitFunc [subst $gamma1a]

      set RhoFunc [subst $gamma2a]

      set TauFunc [subst $gamma3a]

      

      #insert "start bdd manipulations"
      update idletasks
      if {$check==2} {
        #puts "INPUT : $INPUT"
        #set InputTrace_a {}
        #foreach el $INPUT {
        # set InputTrace_a_el [input_to_trace $el]
        # lappend InputTrace_a $InputTrace_a_el
        #}
       
        #puts "InputTrace_a : $InputTrace_a"
        set InputTrace {}
        foreach el $INPUT {
          #puts "el: $el"
          set el_a [input_to_trace $el]
          set el_bdd [subst $el_a]
         # puts "el_bdd: $el_bdd"
          lappend InputTrace $el_bdd
        }
        #puts "InputTrace: $InputTrace"
        set t_enforce [time {enforce $suffix_list $k $InitFunc $RhoFunc $TauFunc $InputTrace}]
        insert "enforce time: $t_enforce" 
        set t_quit [time {Quit-data-manager}]
        insert "quit time: $t_quit" 
      } else {
        if {$check==6} {
        #puts "INPUT : $INPUT"
        #set InputTrace_a {}
        #foreach el $INPUT {
        # set InputTrace_a_el [input_to_trace $el]
        # lappend InputTrace_a $InputTrace_a_el
        #}
       
        #puts "InputTrace_a : $InputTrace_a"
        set InputTrace {}
        foreach el $INPUT {
          #puts "el: $el"
          set el_a [input_to_trace $el]
          set el_bdd [subst $el_a]
         # puts "el_bdd: $el_bdd"
          lappend InputTrace $el_bdd
        }
        #puts "InputTrace: $InputTrace"
        set t_enforce [time {p_enforce $suffix_list $k $InitFunc $RhoFunc $TauFunc $InputTrace}]
        insert "enforce time: $t_enforce" 
        set t_quit [time {Quit-data-manager}]
        insert "quit time: $t_quit"
        } else { 
          if {$check==3} {
            i_enforce $suffix_list $k $InitFunc $RhoFunc $TauFunc
            set t_quit [time {Quit-data-manager}]
            insert "quit time: $t_quit" 
          } else {
            if {$check==7} {
            p_i_enforce $suffix_list $k $InitFunc $RhoFunc $TauFunc
            set t_quit [time {Quit-data-manager}]
            insert "quit time: $t_quit"
            } else {
              #puts "at invariant"
              set t [time {invariant $suffix_list $check $k $InitFunc $RhoFunc $TauFunc}]
              insert "invariant time: $t" 
              set t_quit [time {Quit-data-manager}]
              insert "quit time: $t_quit"
            }
          }
        }
       }
      } else {
        insert "syntax error"
      }
}

}
# end ::bdd_tcl namespace 


# GUI paremeters
#


set cmd {}
set point 0
set event 1
#set bdd_tcl::count 0
set tab 3
set current ""
set BDD "CUDD"
set nodisplay 0



set arglen [llength $argv]
set index 0
 while {$index < $arglen} {
   set arg [lindex $argv $index]
   switch -exact -- $arg {
       {-buddy} {set BDD "Buddy"} 
       {-file}  {::bdd_tcl::check_file "[lindex $argv [incr index]]"}
       {-i_enf_l} {set nodisplay 1; set input_tcl "i_enf_l [lindex $argv [incr index]]"}
       {-i_enf_r} {set nodisplay 1; set input_tcl "i_enf_r [lindex $argv [incr index]]"}
       {-sat_l}   {set nodisplay 1; set input_tcl "sat_l [lindex $argv [incr index]]"}
       {-sat_r}   {set nodisplay 1; set input_tcl "sat_r [lindex $argv [incr index]]"}
       {-vld_l}   {set nodisplay 1; set input_tcl "vld_l [lindex $argv [incr index]]"}
       {-vld_r}   {set nodisplay 1; set input_tcl "vld_r [lindex $argv [incr index]]"}
       {-enf_l}   {set nodisplay 1; set input_tcl "enf_l [lindex $argv [incr index]]"}
       {-enf_r}   {set nodisplay 1; set input_tcl "enf_r [lindex $argv [incr index]]"}
       default  {set nodisplay 0}
   }
   incr index
 }



if [string equal $BDD "CUDD"] {
  # CUDD library
 if {![info exist ::starkit::topdir]} { 
  if [string equal $::tcl_platform(platform) "windows"] { 
    load windows/cudd_tcl.dll
  } else {
    if [string equal $::tcl_platform(os) "SunOS"] {
     puts "Platform not suported"
     exit
    } else {
      if [string equal $::tcl_platform(os) "Linux"] {
        if [string equal $::tcl_platform(machine) "x86_64"] {
          load linux64/cudd_tcl.so
        } else {
          load linux/cudd_tcl.so
        }
      } else {
        if [string equal $::tcl_platform(os) "Darwin"] {
          load macosx/cudd_tcl.dylib
        } else {
          puts "Platform not suported yet"
          exit
        }
      }
    }
  }
 } else {
   package require cudd_tcl
  }
} else {
  if [string equal $BDD "Buddy"] {
    # Buddy library
   if {![info exist ::starkit::topdir]} { 
    if [string equal $::tcl_platform(platform) "windows"] { 
    load windows/buddy_tcl.dll
    } else {
      if [string equal $::tcl_platform(os) "SunOS"] {
       puts "Platform not suported"
       exit
      } else {
        if [string equal $::tcl_platform(os) "Linux"] {
          if [string equal $::tcl_platform(machine) "x86_64"] {
            load linux64/buddy_tcl.so
          } else {
             load linux/buddy_tcl.so
          }
        } else {
          if [string equal $::tcl_platform(os) "Darwin"] {
           load macosx/buddy_tcl.dylib
          } else {
            puts "Platform not suported yet"
            exit
          }
        }
      }
    }
   } else {
    package require buddy_tcl
   }
  } else {
  puts "no bdd selected"
  exit
 }
}

bdd_tcl::init_bdd



if { $nodisplay == 0 } { 
    
  set fore black
  set back #99aecb
  tk_setPalette $back
  set tback white
  set oback white
  set bddfont {Courier 14} 
  ttk::style configure . -font $bddfont
  option add *Text.background $tback widgetDefault

  ttk::setTheme strl
  ::bdd_tcl::Window show .
  ::bdd_tcl::Window show .top18
  $out tag configure blue -foreground blue
  $out tag configure normal -foreground black
  $out tag configure red  -foreground red
  $out tag configure green  -foreground green

  $help tag configure blue -foreground blue
  $help tag configure normal -foreground black
  $help tag configure red  -foreground red
  $help tag configure green  -foreground green

  ::bdd_tcl::insert_help "\nFLCHECK is a decision procedure for Fusion Logic\n" blue
  ::bdd_tcl::insert_help "It is written in Tcl/Tk and uses the CUDD bdd library\[1\] of Fabio
Somenzi and the BuDDy library package of Jorn Lind-Nielsen\[2\]. Swig\[3\]
is used to write the Tcl/Tk interface to both CUDD and Buddy. The
first version of the fusion logic decision procedure was written by
Ben Moszkowski using perldDD\[4\] (perl interface to CUDD) and clisp. 

The formal details of the decision procedure are in \[5,6,7,8\]. The
enforcement technique is described in \[7,8\]. "

  ::bdd_tcl::insert_help "\nSyntax of Fusion Logic\n" blue

  ::bdd_tcl::insert_help "variable   

  v ::= \[A-Z\]+\[_A-Z\]*

state formula  

  w ::= v | true | false | w or w | w and w | { w } | not w

transition formula  
  t ::= v | true | false | t or t | t and T | { t } | next w

fusion expression   
  
  e ::= chopstar e | step t | test w | e or e | e ; e | { e } 

right fusion logic formula 

  r ::= v | true | false | < e > r | r or r | r and r | not r | { r } 

left fusion logic formula

  l ::= v | true | false | l < e > | l or l | l and l | not l | { l } 
  
 

Precedence of operators:
(from lowest to highest)

or (fusion logic formula)
and (fusion logic formula)
or (state formula)
and (state formula)
or (transition formula)
and (transition formula)
or (fusion expression)
< >  (fusion logic chop)
;    (fusion expression chop)
chopstar
test
step
next
not"

  ::bdd_tcl::insert_help "\nUsing TCL syntax to define abbreviations\n" blue

  ::bdd_tcl::insert_help "TCL variables holds as values strings

assignment of string \"not A\" to TCL variable y is as follows

set y \"not A\"

To refer to the value of TCL variable y is as follows

set z \"\$y or B\"

The value of TCL variable z is then

\"not A or B\"

Using TCL syntax to define derived operators

Let f_1 and f_2 be two fusion logic formulae then

`f_1 implies f_2' denotes the derived fusion logic formula

`not { f_1 } or f_2'

The following TCL code 

proc implies {a b} {
    return \"{not{\$a} or \$b}\"
}

defines the derived fusion logic operator `implies'

The derived fusion logic operator `implies' is used as follows

set x \"\[implies A B\]\"

The value of TCL variable x is then

\"{ not { A } or B }\"

The following derived operators are defined
(see fusion_logic_derived.tcl for their definitions)

\[implies \$a \$b\]

\[equiv \$a \$b\]    

\[len_e \$n\] (n an integer)

\[true_e\]

\[sometimes_e\]

\[skip_e\]

\[empty_e\]

\[len_r \$n\] (n an integer)

\[next_r \$a\]

\[sometimes_r \$a\]

\[always_r \$a\]

\[di_r \$a\]

\[bi_r \$a\]

\[more_r\]

\[empty_r\]

\[skip_r\]

\[fin_r \$a\]

\[len_l \$n\] (n an integer)

\[next_l \$a\]

\[sometimes_l \$a\]

\[always_l \$a\]

\[di_l \$a\]

\[bi_l \$a\]

\[more_l\]

\[empty_l\]

\[skip_l\]

\[fin_l \$a\]

\[and \$a \$b\] (for legacy reasons)

\[false\] (for legacy reasons)"

  ::bdd_tcl::insert_help "\nChecking Validity and Satisfiability\n" blue

  ::bdd_tcl::insert_help "check left/right fusion logic formula f for validity

  vld_l f   vld_r f 

check left/right fusion logic formula f for satisfiability

  sat_l f   sat_r f"

  ::bdd_tcl::insert_help "\nChecking Enforcement\n" blue

  ::bdd_tcl::insert_help "enforce left/right fusion logic formula f wrt a given input 

Let TCL variable INPUT be the input for which enforcement of f is
checked  

  enf_l f   enf_r f

interactive enforce left/right fusion logic formula f

  i_enf_l f    i_enf_r f

You will be prompted for the input trace state by state where `end' will
denote end of the input."

  ::bdd_tcl::insert_help "\nCommand line options\n" blue

  ::bdd_tcl::insert_help "flcheck.tcl without command line options will start the graphical user
interface.  In the lower window line of the GUI one can input the
check ones want to perform, for instance, vld {A or B}. The top window
shows the output of the decision procedure while the middle window
shows the history of the checks performed. The history of the checks
performed can be accessed in the lower window line using the up and
down cursor keys.


The following command line options are available:

flcheck.tcl \[-buddy\] -file <filename> \[ -vld_l | -vld_r | -sat_l | -sat_r | -enf_l | -enf_r | -i_enf_l | -i_enf_r \] \"formula\" 

-buddy: use BuDDY bdd package instead of the CUDD bdd library

-file <filename>: use definitions in file <filename>

-vld_l \"formula\": check the validity of left fusion \"formula\" without GUI

-vld_r \"formula\": check the validity of right fusion \"formula\" without GUI

-sat_l \"formula\": check the satisfiability of left fusion \"formula\" without GUI

-sat_r \"formula\": check the satisfiability of right fusion \"formula\" without GUI

-enf_l \"formula\": check the enforcement of left fusion \"formula\" without GUI

-enf_r \"formula\": check the enforcement of right fusion \"formula\" without GUI

-i_enf_l \"formula\": check interactively the enforcement of left fusion \"formula\" without GUI

-i_enf \"formula\": check interactively the enforcement of right fusion \"formula\" without GUI"


  
  ::bdd_tcl::insert_help "\nReferences\n" blue

  ::bdd_tcl::insert_help "\[1\] CUDD library. 
    Fabio Somenzi
    University of Colorado at Boulder
    ECE Dept.
    Boulder, CO 80309-0425
    Fabio@Colorado.EDU
    http://vlsi.colorado.edu/~fabio

\[2\] BuDDy library package.
    Jorn Lind-Nielsen
    http://sourceforge.net/projects/buddy/

\[3\] http://www.swig.org/

\[4\] perldDD.
    Fabio Somenzi
    University of Colorado at Boulder
    ECE Dept.
    Boulder, CO 80309-0425
    Fabio@Colorado.EDU
    http://vlsi.colorado.edu/~fabio

\[5\] @InCollection{Moszkowski05, 
    author = {Ben Moszkowski}, 
     title = {A Hierarchical Analysis of Propositional Temporal Logic based on
             Intervals}, 
 booktitle = {We Will Show Them: Essays in Honour of {Dov Gabbay}}, 
     pages = {371--440}, 
 publisher = {College Publications (formerly KCL Publications)}, 
      year = 2005, 
    editor = {Sergei Artemov and Howard Barringer and 
              Artur S. d'Avila Garcez and Luis C. Lamb and John Woods}, 
    volume = 2, 
   address = {King's College, London},
      isbn = {1-904987-12-5}, 
    mynote = {Publisher's URL: \\url{www.collegepublications.co.uk}, 
              Full title of two-volume set is We Will Show Them: Essays in 
              Honour of {Dov Gabbay} on his 60th Birthday.  
              ISBN of Vol 1 is 1-904987-11-7} 
   }

\[6\] @Article{Moszkowski04a,
  author =       {Ben Moszkowski},
  title =        {A Hierarchical Completeness Proof for {Propositional
                  Interval Temporal Logic} with Finite Time},
  journal =      {Journal of Applied Non-Classical Logics},
  year =         2004,
  volume =       14,
  number =       {1--2},
  pages =        {55--104},
  note =         {Special issue on Interval Temporal Logics and
                  Duration Calculi}
}

\[7\] A. Cau, H. Janicke and B. Moszkowski.
    A Decision Procedure for Fusion Logic. Seminar talk slide, see
    file itl-fl.pdf.

\[8\] A. Cau, H. Janicke and B. Moszkowski.
    Verification and Enforcement of Access Control Policies. 
    Submitted to Formal Methods in System Design journal.

"

  ::bdd_tcl::help_begin

  

  
#  set ttt [file normalize fusion_logic_derived.tcl]
#  ::bdd_tcl::insert "loading $ttt"
  if {![info exist ::starkit::topdir]} {
   source fusion_logic_derived.tcl
  } else {
   source $::starkit::topdir/fusion_logic_derived.tcl
 }

  bind $entry_widget <Return> {+
   ::bdd_tcl::insert_input $enter_text
   ::bdd_tcl::remove_tabs .top18
   ::bdd_tcl::check $enter_text
   set enter_text ""
  }

  bind $entry_widget <Any-Enter> {+
	focus %W
  }

  bind $entry_widget <KeyRelease> {+
	update idletasks
  } 

  bind $notebook_widget <ButtonRelease-1> {+
        update idletasks
  }

} else {
  wm withdraw .
#  set ttt [file normalize fusion_logic_derived.tcl]
#  ::bdd_tcl::insert "loading $ttt"
  if {![info exist ::starkit::topdir]} {
   source fusion_logic_derived.tcl
  } else {
   source $::starkit::topdir/fusion_logic_derived.tcl
  }
  ::bdd_tcl::check $input_tcl
  exit
}
